"""Functions for parsing, cleaning, and normalising data."""

# Importing packages

from ..datasets import stopwords, html_stopwords 

from typing import List, Dict, Tuple
import copy
from datetime import datetime, date, timedelta
import numpy as np
import pandas as pd
from selectolax.parser import HTMLParser # type: ignore
import requests
# import requests_html <-- raising an error
# from requests_html import HTML

import nltk
from nltk.tokenize import word_tokenize, sent_tokenize # type: ignore

nltk.download('punkt')

def join_list_by_colon(item):
    
    if type(item) == list:
        return ', '.join(item)
    
    else:
        return item

def join_df_col_lists_by_colon(dataframe):

    for col in dataframe.columns:
        dataframe[col] = dataframe[col].apply(join_list_by_colon)
    
    return dataframe

def split_str_by_colon(item):
    
    if type(item) == str:
        return item.split(',')
    
    else:
        return item

def join_list_by_semicolon(item: list) -> str:
    
    """
    Joins the items of a list into a string connected by semicolons. 
    """
    
    if type(item) == list:
        
        return '; '.join(item)
    
    else:
        return item

def join_df_col_lists_by_semicolon(dataframe: pd.DataFrame) -> pd.DataFrame:
    
    """
    Replaces the lists in a Dataframe with strings connected by semicolons. 
    """
    
    for col in dataframe.columns:
        dataframe[col] = dataframe[col].apply(join_list_by_semicolon)
    
    return dataframe

def strip_list_str(list_item: List[str]) -> list:
    
    """
    Takes string. If it is formatted as a Python list, splits it and cleans it. Returns it as a list object.
    """
    
    if type(list_item) == str:
        
        list_series = pd.Series(list_item)
        list_series = list_series.str.replace('[', '', regex = False).str.replace(']', '', regex = False).str.replace('"', '', regex = False).str.replace("'", '', regex = False)
        list_series = list_series.replace(np.nan, None).replace('NaN', None).replace('nan', None).replace('none', None).replace('None', None)
        list_series = list_series.str.split(',')
        clean_list = list_series[0]
# 
        return clean_list
    
    else:
        raise TypeError('Item must be a string')

def text_cleaner(data: str, replace: List[str] = ['[', ']', '\n', '\\n', '\\\\']) -> str:
    
    """
    Takes string and removes inputted characters."
    """
    
    for character in replace:
            data = data.replace(character, '')
    
    return data

def remove_stopwords(
                    words_list: List[str], 
                     language: str = 'English', 
                     stopwords_list = 'all', 
                     min_length: int = 2, 
                     max_length: int = 50, 
                     ignore_case: bool = True, 
                     retain_order: bool = False
                    ) -> list:
    
    """
    Takes a sequence of strings and checks if the string is in a list of stopwords. Returns a list object with the strings removed.
    
    Users can select a list of stopwords from the stopwords object. Defaults to 'all' stopwords.
    Users can opt-in to retain the order of the words using the 'retain_order' parameter.
    """
    
    # Importing stopwords object into the environment
    global stopwords
    
    if stopwords_list in stopwords.keys():
        to_remove = stopwords[stopwords_list]
    
    else:
        
        if type(stopwords_list) == list:
            to_remove = stopwords_list
        
        else:
            if (language.lower() == 'english') or (language.lower() == 'en'):
                to_remove = stopwords['en']
    
    # Conducting some data cleaning
    to_remove = pd.Series(to_remove, dtype = object).str.lower().to_list()
    
    words_list_cleaned = pd.Series(words_list, dtype = object)
    words_list_cleaned = words_list_cleaned.str.replace('|', ' ', regex = False).str.replace('/', ' ', regex=False).str.replace('=', ' ', regex = False).str.replace('\n', ' ', regex = False).str.replace('?', ' ', regex=False).str.replace('!', ' ', regex=False).str.replace('“', ' ', regex=False).str.replace('”', ' ', regex=False).str.replace('"', ' ', regex=False).str.replace('"', ' ', regex=False).str.replace('#', ' ', regex=False).str.replace('[', '', regex=False).str.replace(']', '', regex=False).str.replace('(', '', regex=False).str.replace(')', '', regex=False).str.replace('%', ' ', regex=False)
    words_list_cleaned = words_list_cleaned.str.replace("'", '', regex=False).str.replace('\\', ' ', regex=False).str.replace('+', ' ', regex=False).str.replace('=', ' ').str.replace('-', ' ', regex=False).str.replace('/', ' ', regex=False).str.replace('_', ' ', regex=False)
    words_list_cleaned = words_list_cleaned.str.strip(' ').str.strip().str.replace('  ', ' ', regex=False).str.strip()
    
    if ignore_case == True:
        words_list_cleaned = words_list_cleaned.str.lower()
        
    words_list_cleaned = words_list_cleaned[(words_list_cleaned.str.len() >= min_length) & (words_list_cleaned.str.len() <= max_length) & (words_list_cleaned != '') & (words_list_cleaned != None) & (words_list_cleaned != np.nan)]
    words_list_cleaned = words_list_cleaned.to_list()
    
    words_list_joined = ','.join(words_list_cleaned)
    words_list_joined = words_list_joined.replace(' ', ',')
    words_list_cleaned = words_list_joined.split(',')
    
    
    # Converting lists into sets
    remove_words_set = set(to_remove)
    wordslist_set = set(words_list_cleaned)
    
    # Identifying words to be removed and retained
    to_remove = list(wordslist_set.intersection(remove_words_set))
    to_retain = list(wordslist_set.difference(remove_words_set))
    
    # Returning set difference if 'retain_order' parameter is set to False.
    if retain_order == False:
        result = to_retain
    
    else:
    
    # Creating list of strings *not* in the stopwords list. Takes advantage of Pandas' serialisation.
        words_df = pd.DataFrame(words_list_cleaned, columns = ['word'])
        words_df_lowered = words_df.copy(deep=True)
        words_df_lowered['word'] = words_df_lowered['word'].str.lower()
        
        drop_indexes = []
        for word in to_remove:
            indexes_to_remove = list(words_df_lowered[words_df_lowered['word'] == word].index)
            drop_indexes = drop_indexes + indexes_to_remove
            
        words_df = words_df.drop(drop_indexes)
        result_list = words_df['word'].to_list()
        result_joined = ','.join(result_list)
        result_joined = result_joined.replace(' ', ',')
        result = result_joined.split(',')
        
    return result
        
def html_words_cleaner(words_list: List[str], to_remove: List[str] = 'default', ignore_case: bool = True, retain_order: bool = True) -> list:
    
    """
    Takes a sequence of strings and checks if the string is in a list of html stopwords. Returns a list object with the strings removed.
    
    Users can select a list of stopwords from the stopwords object. Defaults to html stopwords.
    Users can opt-in to retain the order of the strings using the 'retain_order' parameter.
    """
    
    # By default, selecting the 'html' stopwords list
    if to_remove == 'default':
        global stopwords
        to_remove = 'html'
    
    # Conducting data cleaning
    words_joined = ','.join(words_list)
    words_joined = words_joined.replace('_', ',').replace('-', ',').replace('^', ',').replace('+', ',').replace('=', ',').replace('.', ',').replace(':', ',').replace('#', ',').replace(';', ',').replace('{', ',').replace('}', ',').replace('(', ',').replace(')', ',').replace('  ', ',').replace(' ', ',')
    split_words = words_joined.split(',')
    
    # Removing stopwords
    output_list = remove_stopwords(split_words, stopwords_list = to_remove, min_length = 2, max_length= 50, ignore_case= ignore_case, retain_order = retain_order)

    return output_list

def html_extract_all_words(html: str = 'request_input', to_remove: List[str] = 'default', ignore_case: bool = True, retain_order: bool = True) -> list:
    
    """
    Takes html as a string object and extracts words. Returns as a list object.
    """
    
    # By default, requests user to input html string
    if html == 'request_input':
        html = input('HTML to parse: ')
    
    # Tokenizing html
    html_tokenized = word_tokenize(html)
    
    # Cleaning html and extracting words
    html_words = html_words_cleaner(words_list = html_tokenized, to_remove = to_remove, ignore_case = ignore_case, retain_order = retain_order)
    
    return html_words

def text_splitter(data: str, parse_by: str = '.', replace: List[str] = ['\n', '\\n', '\\\\']) -> list:
    
    """
    Takes text as a string object, cleans it, and splits it into a set of strings depending on symbols given. Returns as a list object.
    """
    
    if type(data) == str:
        
        for character in replace:
            data = data.replace(character, parse_by)

        data = data.split(parse_by)
        
        split_data = []
        for element in data:
            if element != '':
                element = element.strip()
                split_data.append(element)

                
        return split_data
    
    
    else:
        return
    
def html_get_tags(html: str) -> list:
    
    """
    Takes html as a string object and extracts tags. Returns as a list.
    """
    
    # Cleaning and splitting data
    data = html.replace('>', '>\n')
    data_split = text_splitter(data, parse_by = '\n')
    
    
    # Extracting tags
    tags = []
    for item in data_split:
        
        if len(item) > 0:
            if (item[0] == '<') and (item[-1] == '>'):
                tags.append(item)
    
    
    tags = list(filter(None, tags))
    
    return tags

def get_text_selectolax(html:str) -> str:
    
    """
    Takes html as a string object and returns the text it contains.
    """
    
    
    tree = HTMLParser(html)

    if tree.body is None:
        return None

    for tag in tree.css('script'):
        tag.decompose()
    for tag in tree.css('style'):
        tag.decompose()

    text = tree.body.text(separator='\n')
    
    return text

def html_text_cleaner(text: str) -> str:
    
    """
    Takes html as a string; removes unwanted characters and strings.
    """
    
    def stripwords(item):
    
        item = item.replace('imprint ads', ' ').replace('javascript is not available', ' ').replace('javascript is disabled', ' ').replace('enable javascript', ' ').replace('supported browsers', ' ').replace('supported browser', ' ').replace('cookie policy', ' ').replace('/div', '').replace('layer layer', ' ').replace('function', ' ').replace('cookie', ' ').replace('page path', ' ').replace('true);', ' ').replace('ua-', ' ').replace('a llow', ' ').replace('allow', ' ').replace('d eveloper', ' ').replace('Developer', ' ').replace('developer', ' ').replace('placehol der', ' ').replace('anonymize ip', ' ').replace(';function', ' ').replace('push(arguments)', ' ').replace('datalayer', ' ').replace(':false', ' ').replace('ad personalization', ' ').replace('default', ' ').replace('page location', ' ').replace('placeholder', ' ').replace('config', ' ').replace('window.data', ' ').replace('window.datalayer', ' ').replace('window.dataLayer', ' ').replace('latest news', ' ').replace('Organization Structure', ' ').replace('Agenda', ' ').replace('\0Tweets', ' ').replace('mx-1', ' ').replace(' b ', ' ').replace('EN', ' ').replace('tooltip', ' ').replace('gtag', ' ').replace('id.', ' ').replace(':true', ' ').replace(' js ', ' ').replace('bottom=', ' ').replace('title=', ' ').replace('\\xa0', ' ').replace('xa0', ' ').replace('style=', ' ').replace('page_placeholder', ' ').replace('class=', ' ').replace('toggle', ' ').replace('data-', ' ').replace('a0', ' ').replace('\\x', ' ').replace('href=', ' ').replace("site map", ' ').replace("Sites Map", ' ').replace("sites map", ' ').replace('contact us', ' ').replace('Contact Us', ' ').replace('about us', ' ').replace('About Us', ' ').replace('sitemap', ' ')
        
        return item
    
    
    def stripchars(item):
        
        item = item.replace(' ); ', ' ').replace('); ', ' ').replace(' ) ', ' ').replace(' ( ', ' ').replace(' . ', ' ').replace(' ,', ' ').replace('>', ' ').replace('<', ' ').replace('\\t', ' ').replace('\n', ' ').replace('\\n', ' ').replace('\\\\n', ' ').replace('b\' ', ' ').replace("\'", ' ').replace('\\t', ' ').replace('\n', ' ').replace('\\n', ' ').replace('\\\\n', ' ').replace('b\' ', ' ').replace("\'", ' ').replace('\\', ' ').replace('|', ' ').replace('"', ' ').replace(']', ' ').replace('[', ' ').replace('=', ' ').replace(' ; ', ' ').replace(' ;', ' ').replace(' : ', ' ').replace('{', ' ').replace('}', ' ').replace('{}', ' ').replace('()', ' ').replace('_', ' ').replace(' s ', "'s ")

        return item
    
    def stripspace(item):
        
        item = item.replace('  ',' ').replace('   ', ' ').replace('     ', ' ').replace('       ', ' ')
        item = item.strip()
        
        return item
    
    text = stripwords(text)
    text = stripchars(text)
    text = stripspace(text)
    text = text.lower()
    text = stripwords(text)
    text = stripchars(text)
    text = stripspace(text)
    
    return text

def get_words_list(text: str, min_length = 2, lower = False) -> list:
    
    """
    Takes text as a string and returns a list of words.
    """
    
    # Removing leading and trailing whitespaces
    text = text.strip()
    
    # Replacing punctuation prior to splitting
    text = text.replace(' -', '').replace('!', '').replace('?', '').replace(',', '').replace('.', '').replace('(', '').replace(')', '').replace(':', '').replace(';', '')
    
    # Splitting string into list
    words_list = text.split(' ')
    
    # Removing empty strings and words shorter than the provided minimum length
    words_list_cleaned = [i for i in words_list if ((i != '') and (i != ' ') and (len(i) > min_length))]
    
    return words_list

def get_html_clean_text(html: str) -> str:
    
    """
    Takes html, extracts text, cleans it, and returns it as a string.
    """
    
    text = get_text_selectolax(html = html)
    text = html_text_cleaner(text)
    
    return text

def get_html_clean_words(html: str) -> list:
    
    """
    Takes html, extracts text, cleans it, and returns words contained as a string.
    """
    
    text = get_html_clean_text(html = html)
    words_list = get_words_list(text = text)
    
    return words_list

def list_remove_nones(list_item: list) -> list:
    
    """
    Removes None items from list.
    """
    
    return [i for i in list_item if i != None]

def none_list_to_empty_list(list_item: list) -> list:
    
    """
    Turns a list of None items into an empty list.
    """
    
    # Checking if object is a list
    if type(list_item) != list:
        raise TypeError('Item must be a list')
    
    # Returning empty list if list is a list of Nones
    if (len(list_item) == 1) and ((list_item[0] == None) or (list_item[0].lower() == 'none') or (str(list_item[0]) == 'NaT') or (list_item[0] == np.nan) or (list_item[0] == pd.NaT)):
        return list()
    
    else:
        return list_item

def nat_list_to_nones_list(list_item: list) -> list:
    
    """
    Replaces Numpy Not a Time (NaT) items in a list with Nones.
    """
    
    
    # Checking if object is a list
    if type(list_item) != list:
        raise TypeError('Item must be a list')
    
    
    # Replacing NaTs with Nones
    new_list = []
    
    for i in list_item:
        if type(i) == type(pd.NaT):
            new_list.append(None)
        else:
            new_list.append(i)
            
    # If list is (now) empty, returns a None object
    if (len(list_item) == 1) and (type(list_item[0]) == None):
        return None
    
    return new_list

def series_none_list_to_empty_lists(item):
    
    """
    Replaces None items in Pandas series with empty lists. To be use with pandas.apply().
    """
    
    try:
        return none_list_to_empty_list(item)
    except:
        return item

def empty_list_to_none(list_item: list):
    
    """
    Replaces empty list with None object.
    """
    
    # Checking if item is a list object
    if type(list_item) != list:
        raise TypeError('Item must be a list')
    
    if (len(list_item) == 1) and (list_item[0].lower() == 'none'):
        return None
    
    if len(list_item) == 0:
        return None
    
    else:
        return list_item

def empty_set_to_none(set_item: set):
    
    """
    Replaces empty set with None object.
    """
    
    if type(set_item) != set:
        raise TypeError('Item must be a set')
    
    if len(set_item) == 0:
        return None
    
    else:
        return set_item

def empty_dict_to_none(dict_item: dict):
    
    """
    Replaces empty dictionary with None object.
    """
    
    if type(dict_item) != dict:
        raise TypeError('Item must be a dictionary')
    
    if len(dict_item) == 0:
        return None
    
    else:
        return dict_item

def empty_to_none(item):
    
    """
    Replaces empty object with None object.
    """
    
    if type(item) == list:
        try:
            return empty_list_to_none(item)
        except:
            pass
    
    if type(item) == set:
        try:
            return empty_set_to_none(item)
        except:
            pass
    
    if type(item) == dict:
        try:
            return empty_dict_to_none(item)
        except:
            pass
    
    try:
        if len(item) == 0:
            return None
        else:
            return item
        
    except:
        return item
    
def parse_data(data, data_type: str, ignore_case: bool = True, stopwords: str = 'all', language: str = 'english', retain_word_order: bool = False) -> dict:
    
    """
    Parses string object based on an inputted datatype. Returns dictionary of parsed contents.
    """
    
    # Checking if data is None object. If so, returns an empty dictionary.
    if data == None:
        return {}
    
    # Checking the type of the data
    data_var_type = type(data)
    
    output = None
    
    # Processing data if datatype given is text
    if data_type == 'text':
        
        # Checking type
        if data_var_type != str:
            raise TypeError('Text data must be inputted as a string')
        
        # Lowering case for data cleaning purposes
        if ignore_case == True:
            data = data.lower()
        
        data_no_para = data.replace('\n', '')
        
        # Parsing data
        output = {
                'paragraphs': text_splitter(data, parse_by = '\n'),
                'sentences': sent_tokenize(data_no_para),
                'phrases': text_splitter(data, parse_by = '.', replace = [
                                                                            '\n',
                                                                            '(',
                                                                            ')',
                                                                            '['
                                                                            ']'
                                                                            '}'
                                                                            '{'
                                                                            '\\', 
                                                                            '/', 
                                                                            ',', 
                                                                            ';',
                                                                            ':',
                                                                            ' - ',
                                                                            ' -- ']
                                                                                ),
                'words': word_tokenize(data_no_para)}
#             text_splitter(data, parse_by = ' ', replace = [
#                                                                         '\n',
#                                                                         '(',
#                                                                         ')',
#                                                                         '['
#                                                                         ']'
#                                                                         '}'
#                                                                         '{'
#                                                                         '\\', 
#                                                                         '/', 
#                                                                         ',', 
#                                                                         ';',
#                                                                         ':',
#                                                                         ' - ',
#                                                                         ' -- ']
#                                                                                 ),
#                 }
        
        
    # Processing data if datatype given is html
    if (data_type == 'html') or (data_type == 'web code'):
            
            # Getting text if data is an html requests object
            if type(data) == requests_html.HTMLResponse:
                data = data.text
            
            # Lowering case for data cleaning purposes
            if ignore_case == True:
                data = data.lower()
            
            # Parsing data
            output = {}
            output['tags'] = html_get_tags(data)
            output['text'] = get_html_clean_text(data)
            output['words'] = get_html_clean_words(data)
            output['element_list'] = text_splitter(data, parse_by = '\n')
    
    # Extracting words and removing stopwords
    if output != None:
        if output['words'] != None:
            output['words'] = remove_stopwords(
                                                words_list = output['words'],
                                                language = language,
                                                stopwords_list = stopwords,
                                                ignore_case = ignore_case,
                                                retain_order = retain_word_order
                                              )
    
    return output

def parse_data_to_set(data, data_type: str, ignore_case: bool = True) -> set:
    
    """
    Parses string data object and returns a set.
    """
    
    # Parsing data
    parsed_data = parse_data(data=data, data_type=data_type, ignore_case = ignore_case)
    
    # Creating set from data.
    for key in parsed_data.keys():
        parsed_data[key] = set(parsed_data[key])
    
    return parsed_data

def is_int(string: str) -> bool:
    
    """
    Checks if a string can be converted successfully to an integer.
    """
    
    try:
        int(string)
        return True
    
    except:
        return False

def correct_int(string: str) -> int:
    
    """
    Converts a string to an integer if it can be converted successfully. Uses exception handling to avoid errors.
    """
    
    boolean = is_int(string)
    
    if boolean == True:
        return int(string)
    
    else:
        return

def is_float(string: str) -> bool:
    
    """
    Checks if a string can be converted successfully to a float.
    """
    
    try:
        float(string)
        return True
    
    except:
        return False

def correct_float(string: str):
    
    """
    Converts a string to a float if it can be converted successfully. Uses exception handling to avoid errors.
    """
    
    boolean = is_float(string)
    
    if boolean == True:
        return float(string)
    
    else:
        return

def date_checker(date: str, print_format: bool = False) -> tuple:
    
    """
    Checks if a string is a date. If successful, returns a tuple containing result and the date's format.
    """
    
    # defining formats
    date_formats = [
            '%y',
            '%Y',
            '%m %Y',
            '%m.%y',
            '%m.%Y',
            '%m/%y',
            '%m/%Y',
            '%m-%y',
            '%m-%Y',
            '%Y %m',
            '%y.%m',
            '%Y.%m',
            '%y/%m',
            '%Y/%m',
            '%y-%m',
            '%Y-%m',
            '%d-%m-%y',
            '%d/%m/%y',
            '%d.%m.%y',
            '%d-%m-%Y',
            '%d/%m/%Y',
            '%d.%m.%Y',
            '%m-%d-%y',
            '%m/%d/%y',
            '%m.%d.%y',
            '%m-%d-%Y',
            '%m/%d/%Y',
            '%m.%d.%Y',
            '%y-%m-%d',
            '%y/%m/%d',
            '%y.%m.%d',
            '%Y-%m-%d',
            '%Y/%m/%d',
            '%Y.%m.%d'
          ]
    
    date_format = ''
    
    for i in date_formats:

        # using try-except to check for truth value
        try:
            res = bool(datetime.strptime(date, i))
            if res == True:
                date_format = i
                if print_format == True:
                    print('Date format: ' + i)
                break
            else:
                continue

        except:
            res = False
            continue
    
    return (res, date, date_format)

def is_date(string: str, print_format: bool = False):
    
    """
    Checks if a string is a date.
    """
    
    if type(string) == datetime:
        return (True, string, None)
    
    if type(string) != str:
        raise TypeError('Input must be a string')
        
    string = string.strip()
    
    date_result = date_checker(date = string, print_format = print_format)
    
    if date_result[0] == False:

        if ('T' in string) or ('t' in string) or (': ' in string) or (',' in string) or (' ' in string):
            if ('T' in string) and ('UTC' not in string) and ('utc' not in string):
                string = string.split('T')[0].strip()

            else:
                if ('t' in string):
                    string = string.split('t')[0].strip()

                else:
                    if (': ' in string):
                        string = string.split(': ')[0].strip()
                    
                    else:
                        if ',' in string:
                            split = string.split(',')
                            if len(split) <=2:
                                string = split[0].strip()
                                
                            
                        elif ' ' in string:
                            split = string.split(' ')
                            if len(split) <=2:
                                string = split[0].strip()
                        
        date_result = date_checker(date = string, print_format = print_format)
    
    return date_result

def str_to_date(string: str, print_format: bool = False) -> datetime:
    
    """
    Converts date strings to datetime objects.
    """
    
    if type(string) == datetime:
        return string
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    result = is_date(string, print_format = print_format)
    if result[0] == True:
    
        try:
            return datetime.strptime(result[1], result[2])

        except:
            raise TypeError('Input is not a valid date')
    
    else:
        raise TypeError('Input is not a valid date')

def time_checker(time: str, print_format: bool= False) -> tuple:
    
    """
    Checks if a string is a time.
    """
    
    # defining formats
    time_formats = [
            '%I%p',
            '%H%M',
            '%H %M',
            '%H/%M',
            '%H.%M',
            '%H:%M',
            '%H, %M',
            '%I%p%M',
            '%I%p %M',
            '%I%p, %M',
            '%H/%M/%S',
            '%H.%M.%S',
            '%H:%M:%S',
            '%H, %M, %S',
            '%H %M %S',
            '%I%p%M%S',
            '%I%p/%M/%S',
            '%I%p.%M.%S',
            '%I%p:%M:%S',
            '%I%p, %M, %S',
            '%I%p %M %S'
          ]
    
    time_format = ''
    
    if len(time) > 0:
        
        # Checking string against all time formats given
        for i in time_formats:

            # using try-except to check for truth value
            try:
                time_res = bool(datetime.strptime(time, i))
                if time_res == True:
                    time_format = i
                    if print_format == True:
                        print('Time format: ' + i)
                    break
                    
                else:
                    time_format = ''
                    continue

            except:
                time_format = ''
                time_res = False
                continue
    
    else:
        time_format = ''
        time_res = False
    
    return (time_res, time, time_format)

def utc_checker(utc_offset, print_format = False):
    
    """
    Checks if a string is in UTC datetime format.
    """
    
    # defining formats
    utc_formats = [
            '%z',
            '%Z',
            '%Z %z',
            '%Z%z',
            '%:z'
          ]
    
    utc_format = ''
    
    if len(utc_offset) > 0:
        
        # Checking string against all UTC datetime formats given
        for i in utc_formats:

            # using try-except to check for truth value
            try:
                utc_res = bool(datetime.strptime(utc_offset, i))
                if utc_res == True:
                    utc_format = i
                    if print_format == True:
                        print('UTC format: ' + i)
                    break
                else:
                    utc_format = ''
                    continue

            except:
                utc_format = ''
                utc_res = False
                continue
    
    else:
        utc_format = ''
        utc_res = False
    
    return (utc_res, utc_offset, utc_format)

def is_time(string: str, print_format: bool = False):
    
    """
    Checks if a string is a time, accounting for UTC formatting.
    """
    
    if type(string) == datetime:
        return (True, string, None)
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    # Stripping leading and trailing whitespaces
    string = string.strip()
    
    # Checking if the string is a time using time checker
    time = string
    utc_offset = ''
    time_res = time_checker(time, print_format = print_format)
    
    # If the result is False, checking alternative formats.
    if time_res[0] == False:
        
        if ('T' in string) or ('t' in string) or (': ' in string) or (',' in string) or (' ' in string):
            if ('T' in string) and ('UTC' not in string) and ('utc' not in string):
                string = string.split('T')[1].strip()

            else:
                if ('t' in string):
                    string = string.split('t')[1].strip()

                else:
                    if (': ' in string):
                        string = string.split(': ')[1].strip()
                    
                    else:
                        if ',' in string:
                            split = string.split(',')
                            if len(split) <=2:
                                string = split[1].strip()
                            
                        elif ' ' in string:
                            split = string.split(' ')
                            if len(split) <=2:
                                string = split[1].strip()
    
        time = string
        utc_offset = ''

        time_res = time_checker(time, print_format = print_format)
    
    
    # If the result is False, checking alternative formats.
    if time_res[0] == False:
        
        if ('+' in string) or ('-' in string) or ('utc' in string.lower()):
        
            if ('UTC' in string) and (('+' in string) or ('-' in string)):
                string = string.replace('UTC', '')
            
            if ('UTC' in string):
                string = string.replace('UTC', '+')
            
            if ('utc' in string) and (('+' in string) or ('-' in string)):
                string = string.replace('utc', '')
            
            if ('utc' in string):
                string = string.replace('utc', '+')
            
            if '+' in string:
                    split = string.split('+')
                    time = split[0].strip()
                    utc_offset = '+' + split[1].strip()

            elif ('-' in string):
                    split = string.split('-')
                    time = split[0].strip()
                    utc_offset = '-' + split[1].strip()
        
        else:
            time = string
            utc_offset = ''
    
    
    # If the result is False, checking alternative formats and UTC formatting.
    
    time_res = time_checker(time, print_format = print_format)
    utc_res = utc_checker(utc_offset = utc_offset, print_format = print_format)
    
    if time_res[0] == True:
        new_time_str = time_res[1]
    else:
        new_time_str = time
    
    if utc_res[0] == True:
        new_utc_res = utc_res[1]
    else:
        new_utc_res = utc_offset
    
    new_str = new_time_str + new_utc_res
    new_format = time_res[2] + utc_res[2]
    
    if (time_res[0] == True) or (utc_res[0] == True):
        return (True, new_str, new_format)
    
    else:
        return (False, new_str, new_format)

def str_to_time(string: str, print_format: bool = False):
    
    """
    Converts time strings to time objects.
    """
    
    if type(string) == datetime:
        return string
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    # Checking if the string is a time
    result = is_time(string, print_format = print_format)
    if result[0] == True:
    
        try:
            return datetime.strptime(result[1], result[2])

        except:
            raise TypeError('Input is not a valid time')
    
    else:
        raise TypeError('Input is not a valid time')
      
def is_datetime(string: str, print_format: bool = False) -> tuple:
    
    """
    Checks if string is a date, time, or date-time.
    """
    
    if type(string) == datetime:
        return (True, string, None)
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    input_string = copy.deepcopy(string)
    
    # Checking if the string is a date or time
    date_res = is_date(string, print_format)
    time_res = is_time(string, print_format)
    
    if (date_res[0] == True) and (time_res[0] == False):
        return date_res
    
    if (date_res[0] == False) and (time_res[0] == True):
        return time_res
    
    if (date_res[0] == True) and (time_res[0] == True):
        return date_res
    
    else:
        return (False, input_string, None)
    
    
    
    # If the result is False, checking alternative formats and UTC formatting.
    if (date_res[0] == False) and (time_res[0] == False):
        
        if ('T' in string) or ('t' in string) or (': ' in string) or (' ' in string) or (',' in string):

            if ('T' in string):
                split = string.split('T')
                date_string = split[0].strip()
                time_string = split[1].strip()

            else:
                if ('T' in string):
                    split = string.split('T')
                    date_string = split[0].strip()
                    time_string = split[1].strip()

                else:
                    
                    
                    if (',' in string):
                        split = string.split(',')
                        date_string = split[0].strip()
                        time_string = split[1].strip()

                    else:
                        
                        if (': ' in string):
                            split = string.split(' ')
                            date_string = split[0].strip()
                            time_string = split[1].strip()
                        
                        elif (' ' in string):
                            split = string.split(' ')
                            date_string = split[0].strip()
                            time_string = split[1].strip()

        else:
            date_string = string
            time_string = ''

        date_res = is_date(date_string, print_format)
        time_res = is_time(time_string, print_format)
    
    
    
    
    if (date_res[0] == True) and (time_res[0] == True):
        new_str = date_res[1] + ' ' + time_res[1]
        new_format = date_res[2] + ' ' + time_res[2]
        return (True, new_str, new_format)
    
    else:
        return (False, input_string, None)

def is_date_or_time(string: str, print_format: bool = False) -> bool:
    
    """
    Checks if string is a date or time.
    """
    
    if type(string) == datetime:
        return (True, string, None)
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    dt_res = is_datetime(string, print_format)
    
    if dt_res == False:
        date_res = is_date(string, print_format)[0]
        time_res = is_time(string, print_format)
    else:
        date_res = True
        time_res = True
    
    if print_format == True:
        print('\nIs date-time: ' + str(dt_res) + '\nContains date: ' + str(date_res) + '\nContains time: ' + str(time_res))
        
    if (date_res == True) or (time_res == True) or (dt_res == True):
        return True
    else:
        return False

def str_to_datetime(string: str, print_format:bool = False) -> datetime:
    
    """
    Converts date, time, or datetime strings to datetime objects.
    """
    
    if type(string) == datetime:
        return string
    
    if type(string) != str:
        raise TypeError('Input must be a string')
    
    # Checking if the item is a datetime
    dt_result = is_datetime(string, print_format = print_format)
    
    if dt_result[0] == True:
        return datetime.strptime(dt_result[1], dt_result[2])
    
    else:
        date_result = is_date(string, print_format = print_format)
        time_result = is_time(string, print_format = print_format)
    
        if (date_result[0] == True) and (time_result[0] == True):
            new_str = date_result[1] + ' ' + time_result[1]
            new_format = date_result[2] + ' ' + time_result[2]
            try:
                return datetime.strptime(new_str, new_format)

            except:
                raise TypeError('Input is not a valid date-time')
    
        else:
            if date_result[0] == True:
                
                try:
                    return datetime.strptime(date_result[1], date_result[2])

                except:
                    raise TypeError('Input is not a valid date-time')

            else:
                if time_result[0] == True:
                    try:
                        return datetime.strptime(time_result[1], time_result[2])

                    except:
                        raise TypeError('Input is not a valid date-time')
                
                else:
                    raise TypeError('Input is not a valid date-time')

def list_to_datetimes(item: list) -> list:
    
    """
    Converts a list of date or time strings into a list of datetime objects.
    """
    
    if (type(item) == list):
        item = pd.to_datetime(pd.Series(item, dtype=object)).to_list()
        return item
        
    else:
        return item

def series_to_datetimes(item: pd.Series) -> pd.Series:
    
    """
    Converts a Pandas series of dates into a series of datetime objects.
    """
    
    if type(item) == str:
        try:
            return str_to_datetime(item, False)
        except:
            return item
    
    else:
        if type(item) == list:
            try:
                return pd.to_datetime(item.astype(object))
            except:
                return item
        
        else: 
            return item

def correct_series_of_lists(series: pd.Series) -> pd.Series:
    
    """
    Converts a Pandas series of strings representing lists into a series of lists.
    """
    
    if type(series) != pd.Series:
        raise TypeError('Series must be a pandas Series object')
    
    index = list(series.index)
    for i in index:
        
        item = series.loc[i]
        
        if type(item) == str:
            item = item.replace('[', '').replace(']', '').replace('{', '').replace('}', '').replace('"', '').replace("'", "").strip().split(',')
            item = [i.strip() for i in item]

        series.at[i] = item
    
    return series

def merge_duplicate_ids(dataframe, merge_on: str):

    df = dataframe.copy(deep=True)
    
    if merge_on in df.columns:

        # sorting index to prioritise entries with real values

        df = df.sort_values(merge_on)

        # Checking if there are valid IDs in the column (i.e., not None or NaN)
        dropna_indexes = df[merge_on].dropna().index

        if len(dropna_indexes) > 0:

            df_dropna = df.copy(deep=True).loc[dropna_indexes]

            try:
                ids = set(df_dropna[merge_on].to_list())
            except:
                ids = set(df_dropna[merge_on].astype(str).to_list())

            for i in ids:

                masked = df[df[merge_on]==i]
                masked_indexes = masked.index.to_list()
                masked = masked.reset_index().drop('index', axis=1)
                len_masked = len(masked)

                if len_masked > 0:

                    first_index = masked_indexes[0]
                    duplicate_indexes = masked_indexes[1:]
                    first_row = masked.loc[0]

                    for index in range(1, len_masked):

                        row = masked.loc[index]

                        for c in first_row.index:

                            data = first_row[c]
                            dtype = type(data)
                            dtype_str = str(dtype)
                            str_data = str(data)

                            if ((data is None) 
                                or (data is np.nan) 
                                or ((dtype == str) and (data == '')) 
                                or ((dtype == str) and (data == '[]'))
                                or ((dtype == str) and (data == '[]'))
                                or  ((dtype == str) and (data == 'None'))
                                or  ((dtype == str) and (data == 'none'))
                                or ((dtype == list) and (data == []))
                                or (str_data == '[]')
                                or (str_data == '{}')
                                ):
                                    first_row[c] = row[c]
                                    continue
                            
                            else:
                                if '.Results' in dtype_str:
                                    data2 = row[c]
                                    dtype2 = type(data2)
                                    dtype_str2 = str(dtype2)
                                    if (dtype2 == pd.DataFrame) or ('.Results' in dtype_str2):
                                        data2_copy = data2.copy(deep=True)
                                        first_row[c].add_dataframe(data2_copy)
                                        first_row[c].remove_duplicates()
                                    continue

                                if ('.Authors' in dtype_str) or ('.Funders' in dtype_str) or ('.Affiliations' in dtype_str):
                                    data2 = row[c]
                                    dtype2 = type(data2)
                                    dtype_str2 = str(dtype2)

                                    if ('.Authors' in dtype_str2) or ('.Funders' in dtype_str2) or ('.Affiliations' in dtype_str2):
                                        concat_df = pd.concat([data.all, data2.all])
                                        concat_df = deduplicate(concat_df)
                                        first_row[c].all = concat_df
                                    continue
  
                    try:
                        first_row = pd.Series(first_row)
                        df.loc[first_index] = first_row
                    except:
                        pass

                    df = df.drop(labels=duplicate_indexes)
        
        dataframe = df

    return dataframe

def merge_all_duplicate_ids(dataframe):

    id_names = ['doi', 'isbn', 'issn', 'uri', 'orcid', 'crossref_id', 'crossref', 'scopus_id', 'scopus', 'wos_id', 'wos', 'pubmed_id', 'address', 'link', 'website']
    df = dataframe.copy(deep=True)



    for i in id_names:
        if i in dataframe.columns:
            dataframe = merge_duplicate_ids(dataframe, merge_on = i)
    
    return dataframe

def deduplicate(dataframe):

    ignore_cols = ['work_id',
                   'author_id',
                   'funder_id',
                   'affiliation_id',
                   'alt_names',
                   'authors', 
                    'funder',
                    'given_name',
                    'affiliations',
                    'citations',
                    'citations_data',
                    'publications',
                    'address',
                    'work_count',
                    'tokens',
                    'other_links',
                    'website',
                    'link',
                    'other_links']

    df = dataframe.copy(deep=True)
    
    if 'doi' in df.columns:
        df['doi'] = df['doi'].str.replace('https://', '', regex = False).str.replace('http://', '', regex = False).str.replace('dx.', '', regex = False).str.replace('doi.org/', '', regex = False).str.replace('doi/', '', regex = False)

    # Creating dataframe without empty columns; converting to string to avoid errors
    df_dropna = df.dropna(axis=1).astype(str)

    # Converting strings to lowercase to improve matching
    for c in df_dropna.columns:
        df_dropna[c] = df_dropna[c].str.lower()

    # Removing duplicates
    col_subset = [c for c in df_dropna.columns if c not in ignore_cols]
    if len(col_subset) > 0:
        first_dedpublication = df_dropna.drop_duplicates(subset=col_subset)
    else:
        first_dedpublication = df_dropna
        
    first_dedpublication_index = first_dedpublication.index

    df2 = dataframe.loc[first_dedpublication_index]

    # Checking for duplicate UIDs; merging rows that share UIDs
    df3 = merge_all_duplicate_ids(df2)

    final_df = df3.reset_index().drop('index', axis=1)

    return final_df


"""
Useful datasets for reference and analysis

Notes
-----
Corpus extracted from names_dataset module + NLTK module. Stored locally for efficiency. See https://pypi.org/project/names-dataset/ for source of first and last names; nltk.corpus.names for source of nltk names
"""
import json
from pathlib import Path

here = str(Path(__file__).parent)

with open(f'{here}/names/all_personal_names.txt', 'r', encoding='utf-8') as file:
    all_personal_names = file.read()
    file.close()
all_personal_names = all_personal_names.replace("'", "").split(', ')

with open(f'{here}/names/first_names.txt', 'r', encoding='utf-8') as file:
    first_names = file.read()
    file.close()
first_names = first_names.replace("'", "").split(', ')

with open(f'{here}/names/last_names.txt', 'r', encoding='utf-8') as file:
    last_names = file.read()
    file.close()
last_names = last_names.replace("'", "").split(', ')

with open(f'{here}/names/nltk_names.txt', 'r', encoding='ascii') as file:
    nltk_names = file.read()
    file.close()
nltk_names = nltk_names.replace("'", "").split(', ')

# Corpus extracted from country_list module. Stored locally for efficiency.

with open(f'{here}/countries/countries_all.txt', 'r', encoding='utf-8') as file:
    countries_all = file.read()
    file.close()
countries_all = countries_all.replace("'", "").split(', ')

with open(f'{here}/countries/country_names.json', 'r', encoding='utf-8') as file:
    country_names = json.load(file)
    file.close()

# Corpus extracted from geonamescache module. Stored locally for efficiency.
with open(f'{here}/cities/cities_all.txt', 'r', encoding='utf-8') as file:
    cities_all = file.read()
    file.close()
cities_all = cities_all.replace("'", "").split(', ')

with open(f'{here}/cities/cities_en.json', 'r', encoding='ascii') as file:
    cities_en = json.load(file)
    file.close()

# Corpus extracted from language_data and langcodes modules. Stored locally for efficiency.

with open(f'{here}/languages/language_names.json', 'r', encoding='ascii') as file:
    language_names = json.load(file)
    file.close()

with open(f'{here}/languages/languages_en.json', 'r', encoding='ascii') as file:
    languages_en = json.load(file)
    file.close()

with open(f'{here}/languages/language_codes.txt', 'r', encoding='ascii') as file:
    language_codes = file.read()
    file.close()

from .stopwords.stopwords import stopwords_dict as stopwords
html_stopwords = stopwords['html']
en_stopwords = stopwords['en']

# Corpus extracted from language_data and langcodes modules. Stored locally for efficiency.

from nltk import download


# Importing NLTK's Word Lists corpus as an NLTK text

try:
    from nltk.corpus import words as nltk_wordlists
    nltk_wordlists.words()
except:
    download('words')
    from nltk.corpus import words as nltk_wordlists


# Importing Swadesh corpus as an NLTK text

try:
    from nltk.corpus import swadesh as nltk_swadesh
    nltk_swadesh.words()
except:
    download('swadesh')
    from nltk.corpus import swadesh as nltk_swadesh

    
# Importing NLTK's Web Text corpus
try:
    from nltk.corpus import webtext as nltk_webtext
    nltk_webtext.words()
except:
    download('webtext')
    from nltk.corpus import webtext as nltk_webtext
      
    
# Importing WordNet 3.0 corpus

try:
    from nltk.corpus import wordnet as nltk_wordnet
    nltk_wordnet.words()
except:
    download('wordnet')
    from nltk.corpus import wordnet as nltk_wordnet

# Creating useful dictionaries of countries in major languages

countries_zh = country_names['zh']
countries_ar = country_names['ar']
countries_es = country_names['es']
countries_hi = country_names['hi']
countries_pt = country_names['pt']
countries_ru = country_names['ru']
countries_fr = country_names['fr']

languages_all = []
for l in language_names.values():
    languages_all = languages_all + list(l.values())

languages_all = list(set(languages_all))


# Extracting NLTK Words Lists as a list of words
nltk_words_list = nltk_wordlists.words()

# Extracting NLTK Web Text  as a list of words
nltk_webtext_words = nltk_webtext.words()



from ..utils.cleaners import deduplicate
from ..importers.orcid import lookup_orcid, get_author, get_author_works
from ..importers.orcid import search as search_orcid # type: ignore
from .entities import Entity, Entities
from .results import Results
from .affiliations import Affiliation, Affiliations, format_affiliations

import pandas as pd
import numpy as np
from pyorcid import Orcid # type: ignore

author_cols = ['author_id',
                                'full_name',
                                'given_name',
                                'family_name',
                                'email',
                                'keywords',
                                'affiliations',
                                'publications',
                                'orcid',
                                'google_scholar',
                                'scopus',
                                'wos',
                                'crossref',
                                'other_links',
                                'other_ids'
                                ]

def get_full_name(series: pd.Series):

            given = series.loc['given_name']
            family = series.loc['family_name']

            if (given is None) or (given is np.nan):
                given = ''
            
            if (family is None) or (family is np.nan):
                family = ''

            if ((type(family) == str) and (',' in family)) and ((given == None) or (given == 'None') or (given == '')):
                split_name = family.split(',')
                given = split_name[0].strip()
                series.loc['given_name'] = given

                family = split_name[1].strip()
                series.loc['family_name'] = family

            full = given + ' ' + family # type: ignore
            full = full.strip()

            if (full == '') or (full == ' '):
                full = 'no_name_given'

            full_name = series.loc['full_name']
            if (full_name == None) or (full_name == 'None') or (full_name == '') or (full_name == 'no_name_given'):
                result = full
            else:
                result = full_name

            return result

def generate_author_id(author_data: pd.Series):

        author_data = author_data.copy(deep=True).dropna().astype(str).str.lower()

        author_id = 'A:'

        if 'given_name' in author_data.index:
            given_name = author_data['given_name']
        else:
            given_name = ''
        
        if 'family_name' in author_data.index:
            family_name = author_data['family_name']
        else:
            family_name = ''

        if 'full_name' in author_data.index:
            full_name = author_data['full_name']
        else:
            full_name = ''

        if ((family_name == None) or (family_name == '') or (family_name == 'None')) and ((full_name != None) and (len(full_name)>0)):
            
            if full_name == 'no_name_given':
                author_id = author_id + '#NA#'
            
            else:
                

                if ' ' in full_name:

                    full_split = full_name.lower().split(' ')
                    first = full_split[0].split(' ')[0].strip()
                    last = full_split[-1]
                
                else:
                    if ',' in full_name:
                        full_split = full_name.lower().split(',')
                        first = full_split[-1].split(' ')[0].strip()
                        last = full_split[0].strip()
                    else:
                        first = full_name
                        last = ''


                author_id = author_id + '-' + first + '-' + last

        else:

            if (given_name != None) and (len(given_name)>0):
                first = given_name.lower().split(' ')[0].strip()
                author_id = author_id + '-' + first
            
            
            if (family_name != None) and (len(family_name)>0):
                family_clean = family_name.lower().strip().replace(' ', '-')
                author_id = author_id + '-' + family_clean

        uid = ''

        if 'orcid' in author_data.index:
            uid = author_data['orcid']
        
        if (uid == None) or (uid == 'None') or (uid == ''):
            
            if 'google_scholar' in author_data.index:
                uid = author_data['google_scholar']
            
            if (uid == None) or (uid == 'None') or (uid == ''):
                if 'scopus' in author_data.index:
                    uid = author_data['scopus']
                
                if (uid == None) or (uid == 'None') or (uid == ''):
                    if 'crossref' in author_data.index:
                        uid = author_data['crossref']
                    
                    if (uid == None) or (uid == 'None') or (uid == ''):
                            uid = ''
        
        uid_shortened = uid.replace('https://', '').replace('http://', '').replace('www.', '').replace('orcid.org/','').replace('scholar.google.com/','').replace('citations?','').replace('user=','')[:30]

        author_id = author_id + '-' + uid_shortened
        author_id = author_id.replace('A:-', 'A:').replace("'s", '').replace('\r', '').replace('\n', '').replace('.', '').replace('(','').replace(')','').replace("'", "").replace('"', '').replace('`','').replace('’','').replace('--', '-').replace('A:-', 'A:').strip('-')

        return author_id

class Author(Entity):

    """
    This is an Author object. It is designed to store data about an individual author and their publications.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self,
                 author_id: str = None, # type: ignore
                 full_name = None, # type: ignore
                 given_name = None, # type: ignore
                 family_name = None, # type: ignore
                 email: str = None, # type: ignore
                 affiliations: str = None, # type: ignore
                 publications: str = None, # type: ignore
                 orcid: str = None, # type: ignore
                 google_scholar: str = None, # type: ignore
                 scopus: str = None, # type: ignore
                 crossref: str = None, # type: ignore
                 other_links: str = None # type: ignore
                 ):
        
        """
        Initialises Author instance.
        
        Parameters
        ----------
        """

        super().__init__()

        if type(given_name) == str:
            given_name = given_name.strip()

        if type(family_name) == str:
            family_name = family_name.strip()

        if ((type(family_name) == str) and (',' in family_name)) and ((given_name == None) or (given_name == '')):
            split_name = family_name.split(',')
            given_name = split_name[0].strip()
            family_name = split_name[1].strip()

        
        global author_cols

        self.summary = pd.DataFrame(columns = author_cols,
                                dtype = object)
        
        
        self.summary.loc[0] = pd.Series(dtype=object)
        self.summary.loc[0, 'author_id'] = author_id
        self.summary.loc[0, 'full_name'] = full_name
        self.summary.loc[0, 'given_name'] = given_name
        self.summary.loc[0, 'family_name'] = family_name
        self.summary.loc[0, 'email'] = email
        self.summary.loc[0, 'affiliations'] = affiliations
        self.summary.loc[0, 'publications'] = publications
        self.summary.loc[0, 'orcid'] = orcid
        self.summary.loc[0, 'google_scholar'] = google_scholar
        self.summary.loc[0, 'scopus'] = scopus
        self.summary.loc[0, 'crossref'] = crossref
        self.summary.loc[0, 'other_links'] = other_links

        full_name = self.get_full_name()
        if full_name != self.summary.loc[0, 'full_name']:
            self.summary.loc[0, 'full_name'] = full_name

        self.publications = Results()

    def generate_id(self):

        author_data = self.summary.loc[0]

        author_id = generate_author_id(author_data) # type: ignore
        return author_id

    def update_id(self):

        current_id = self.summary.loc[0, 'author_id']
        new_id = self.generate_id()

        if (current_id != new_id) or (current_id == None) or (current_id == 'None') or (current_id == '') or (current_id == 'A:#NA#'):
            self.summary.loc[0, 'author_id'] = new_id
        
    
    def __getitem__(self, key):
        
        """
        Retrieves Author attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.summary.columns:
            return self.summary.loc[0, key]
        
        if key in self.publications.columns:
            return self.publications[key]

    def __repr__(self) -> str:
        return str(self.summary.loc[0, 'full_name'])
    
    def get_full_name(self):
        series = self.summary.loc[0]
        return get_full_name(series=series) # type: ignore


    def update_full_name(self):

            full_name = self.get_full_name()
            self.summary.loc[0, 'full_name'] = full_name

    def name_set(self) -> set:

        given = str(self.summary.loc[0, 'given_name'])
        family = str(self.summary.loc[0, 'family_name'])

        return set([given, family])

    def has_orcid(self) -> bool:

        orcid = self.summary.loc[0, 'orcid']

        if (type(orcid) == str) and (orcid != ''):
            return True
        else:
            return False

    def format_affiliations(self):

        affils_data = self.summary.loc[0, 'affiliations']
        affiliations = format_affiliations(affils_data)
        self.summary.at[0, 'affiliations'] = affiliations

    def add_series(self, series: pd.Series):
        self.summary.loc[0] = series

    def from_series(series: pd.Series): # type: ignore
        author = Author()
        author.add_series(series)
        return author
    
    def add_dataframe(self, dataframe: pd.DataFrame):
        series = dataframe.loc[0]
        self.add_series(series) # type: ignore

    def from_dataframe(dataframe: pd.DataFrame): # type: ignore
        author = Author()
        author.add_dataframe(dataframe)
        return author

    def import_crossref(self, crossref_result: dict):

        if 'given' in crossref_result.keys():
            self.summary.loc[0, 'given_name'] = crossref_result['given']
        
        if 'family' in crossref_result.keys():
            self.summary.loc[0, 'family_name'] = crossref_result['family']
        
        if 'email' in crossref_result.keys():
            self.summary.loc[0, 'email'] = crossref_result['email']

        if 'affiliation' in crossref_result.keys():
            if (type(crossref_result['affiliation']) == list) and (len(crossref_result['affiliation']) > 0):
                self.summary.at[0, 'affiliations'] = crossref_result['affiliation'][0]

            else:
                if (type(crossref_result['affiliation']) == dict) and (len(crossref_result['affiliation'].keys()) > 0):
                    key = list(crossref_result['affiliation'].keys())[0]
                    self.summary.at[0, 'affiliations'] = crossref_result['affiliation'][key]

        if 'ORCID' in crossref_result.keys():
            self.summary.loc[0, 'orcid'] = crossref_result['ORCID']
        
        else:
            if 'orcid' in crossref_result.keys():
                self.summary.loc[0, 'orcid'] = crossref_result['orcid']

        # self.summary.loc[0, 'google_scholar'] = google_scholar
        # self.summary.loc[0, 'crossref'] = crossref
        # self.summary.loc[0, 'other_links'] = other_links

        self.update_full_name()
    
    def import_orcid(self, orcid_id: str):

        try:
            auth_res = get_author(orcid_id)
            auth_record = auth_res.record()

        except:
            auth_record = {}
            auth_res = None

            try:
                auth_df = lookup_orcid(orcid_id)
                cols = auth_df.columns.to_list()

                if len(auth_df) > 0:

                    author_details = auth_df.loc[0]

                    if 'name' in cols:
                        self.summary.loc[0, 'given_name'] = author_details['name']
                    
                    if 'family name' in cols:
                        self.summary.loc[0, 'family_name'] = author_details['family name']
                    
                    if 'emails' in cols:
                        self.summary.at[0, 'email'] = author_details['emails']
                    
                    if 'employment' in cols:
                        self.summary.at[0, 'affiliations'] = author_details['employment']
                    
                    if 'works' in cols:
                        self.summary.at[0, 'publications'] = author_details['works']
                    
                    self.summary.loc[0, 'orcid'] = orcid_id
                    self.summary.loc[0, 'orcid'] = orcid_id
                    self.update_full_name()

                    return

            except:
                pass


        if 'person' in auth_record.keys():
            author_details = auth_record['person']
        else:
            author_details = {}
        
        details_keys = author_details.keys()
        
        if 'name' in details_keys:
            
            if 'given-names' in author_details['name']:
                given_list = list(author_details['name']['given-names'].values())
                given = ' '.join(given_list)
                self.summary.loc[0, 'given_name'] = given

            if 'family-name' in author_details['name']:
                family_list = list(author_details['name']['family-name'].values())
                family = ' '.join(family_list)
                self.summary.loc[0, 'family_name'] = family
            
        if 'emails' in details_keys:
            emails_dict = author_details['emails']
            if type(emails_dict) == dict:
                if 'email' in emails_dict.keys():
                    emails_data = emails_dict['email']
                    emails_list = []

                    if (type(emails_data) == list) and (len(emails_data)>0):
                        for i in emails_data:
                            email_addr = i['email']
                            emails_list.append(email_addr)

                    self.summary.at[0, 'email'] = emails_list
        
        if 'keywords' in details_keys:
            kws_dict = author_details['keywords']
            if type(kws_dict) == dict:
                if 'keyword' in kws_dict.keys():
                    kws_data = kws_dict['keyword']
                    kws_list = []

                    if (type(kws_data) == list) and (len(kws_data)>0):
                        for i in kws_data:
                            kwd = i['content']
                            kws_list.append(kwd)

                    self.summary.at[0, 'keywords'] = kws_list

        if 'external-identifiers' in details_keys:
            ext_ids_dict = author_details['external-identifiers']
            if type(ext_ids_dict) == dict:
                if 'external-identifier' in ext_ids_dict.keys():
                    ext_ids_data = ext_ids_dict['external-identifier']
                    ext_ids_formatted = {}

                    if (type(ext_ids_data) == list) and (len(ext_ids_data)>0):
                        for i in ext_ids_data:
                            id_type = i['external-id-type']
                            value = i['external-id-value']
                            url = i['external-id-url']
                            ext_ids_formatted[id_type] = {'id': value, 'url': url}

                    self.summary.at[0, 'keywords'] = ext_ids_formatted

        
        
        self.summary.loc[0, 'orcid'] = orcid_id

        if type(auth_res) == Orcid:
            try:
                works = auth_res.works() # type: ignore
            except:
                works = tuple()
        else:
            works = tuple()
        
        if len(works) > 0:

            works_list = works[0]

            df = pd.DataFrame(works_list)
            df = df.rename(columns={'publication-date': 'date', 'journal title':'source', 'url':'link'})
            df = df.drop(['organization', 'organization-address'], axis=1)

            self.publications.add_dataframe(df)
            self.publications.drop_empty_rows()
            self.publications.remove_duplicates(drop_empty_rows=True)
        
        self.update_full_name()
        

    def from_crossref(crossref_result: dict): # type: ignore

        author = Author()
        author.import_crossref(crossref_result)
        author.update_full_name()

        return author
    
    def from_orcid(orcid_id: str): # type: ignore

        author = Author()
        author.import_orcid(orcid_id)
        author.update_full_name()

        return author
    
    def update_from_orcid(self):

        orcid = self.summary.loc[0, 'orcid']

        if (orcid != None) and (orcid != '') and (orcid != 'None'):
            
            orcid = str(orcid).replace('https://', '').replace('http://', '').replace('orcid.org/', '')

            self.import_orcid(orcid_id = orcid)

class Authors(Entities):

    """
    This is an Authors object. It contains a collection of Authors objects and compiles data about them.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self, authors_data = None):
        
        """
        Initialises Authors instance.
        
        Parameters
        ----------
        """

        super().__init__()
        
        global author_cols
        self.summary = pd.DataFrame(columns = author_cols,
                                dtype = object)
        

        self.all = dict()

        self.data = []
        self.data.append(authors_data)

        if (type(authors_data) == list) and (type(authors_data[0]) == Author):

            for i in authors_data:
                auth = i.summary.copy(deep=True)
                self.summary = pd.concat([self.summary, auth])

            self.summary = self.summary.reset_index().drop('index',axis=1)

        else:

            if type(authors_data) == dict:
                
                values = list(authors_data.values())

                if type(values[0]) == Author:

                    for a in authors_data.keys():
                        
                        index = len(self.summary)
                        auth = a.summary.copy(deep=True)
                        self.summary = pd.concat([self.summary, auth])
                        self.summary.loc[index, 'author_id'] = a

                    self.summary = self.summary.reset_index().drop('index',axis=1)
                


    def __getitem__(self, key):
        
        """
        Retrieves Authors attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.all.keys():
            return self.all[key]

        if key in self.summary.columns:
            return self.summary[key]
        
        if (type(key) == int) and (key <= len(self.data)):
            return self.data[key]
    
    def __repr__(self) -> str:

        alphabetical = self.summary['full_name'].sort_values().to_list().__repr__()
        return alphabetical
    
    def __len__(self) -> int:
        return len(self.all.keys())

    def remove_duplicates(self, drop_empty_rows = True, sync=True):

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        df = self.summary.copy(deep=True)
        df['orcid'] = df['orcid'].str.replace('http://', '', regex=False).str.replace('https://', '', regex=False).str.replace('orcid.org/', '', regex=False).str.strip('/')
        df['google_scholar'] = df['google_scholar'].str.replace('http://', '', regex=False).str.replace('https://', '', regex=False).str.replace('scholar.google.com/', '', regex=False).str.replace('citations?', '', regex=False).str.replace('user=', '', regex=False).str.strip('/')
       
        df = df.sort_values(by = ['orcid', 'google_scholar', 'crossref', 'full_name']).reset_index().drop('index', axis=1)
        self.summary = deduplicate(self.summary)

        if sync == True:
            self.sync_details()

        return self
        


    def merge(self, authors, drop_duplicates = False, drop_empty_rows=False):

        left = self.summary.copy(deep=True)
        right = authors.summary.copy(deep=True)
        
        merged = pd.concat([left, right])

        self.summary = merged.drop_duplicates(subset=['author_id', 'family_name', 'orcid'], ignore_index=True)

        for i in authors.all.keys():
            if i not in self.all.keys():
                self.all[i] = authors.all[i]

        left_data = self.data
        right_data = authors.data

        if left_data == None:
                left_data = []
            
        if right_data == None:
                right_data = []

        if (type(left_data) == Author) or (type(left_data) == str):
                left_data = [left_data]
            
        if (type(right_data) == Author) or (type(right_data) == str):
                right_data = [right_data]
            
        if type(left_data) == dict:
                left_data = list(left_data.values())
            
        if type(right_data) == dict:
                right_data = list(right_data.values())

        merged_data = left_data + right_data # type: ignore
        merged_data = pd.Series(merged_data).value_counts().index.to_list()

        self.data = merged_data

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        return self

    def update_full_names(self):

        for i in self.summary.index:
            new = get_full_name(self.summary.loc[i])
            old = self.summary.loc[i, 'full_name']

            if (type(old) != str) or ((type(old) == str) and (len(old) == 0)) or (old is None) or (old == '') or (old is np.nan):
                self.summary.loc[i, 'full_name'] = new
        
        self.update_author_ids()
        self.sync_details()

    def add_author(self, author: Author, data = None, drop_duplicates = False, drop_empty_rows = False, update_from_orcid = False):

        if update_from_orcid == True:
            orcid = author.summary.loc[0,'orcid']
            if (orcid != None) and (orcid != '') and (orcid != 'None'):
                author.update_from_orcid()

        author.update_id()

        author_id = str(author.summary.loc[0, 'author_id'])

        if (author_id in self.all.keys()) or (author_id in self.summary['author_id'].to_list()):
            print(f'Warning: {author_id} is already in authors')

        # if author_id in self.summary['author_id'].to_list():
        #     id_count = len(self.summary[self.summary['author_id'].str.contains(author_id)]) # type: ignore
        #     author_id = author_id + f'#{id_count + 1}'
        #     author.summary.loc[0, 'author_id'] = author_id

        self.summary = pd.concat([self.summary, author.summary])
        self.summary = self.summary.reset_index().drop('index', axis=1)

        self.all[author_id] = author

        if data == None:
            data = author.summary.to_dict(orient='index')
        
        self.data.append(data)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)


    def add_authors_list(self, authors_list: list, drop_duplicates = False, drop_empty_rows = False):
        
        for i in authors_list:
            if type(i) == Author:
                self.add_author(author = i, drop_duplicates=False)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def mask_entities(self, column, query: str = 'request_input', ignore_case: bool = True):

        if query == 'request_input':
            query = input('Search query').strip()

        query = str(query).strip()

        def entity_masker(entities):
            
            if 'contains' in entities.__dir__():
                res = entities.contains(query=query, ignore_case=ignore_case)
                
            else:
                res = False

            return res
        
        masked = self.summary[self.summary[column].apply(entity_masker)]

        return masked

    def update_author_ids(self):

        for i in self.summary.index:
            author_data = self.summary.loc[i]
            author_id = generate_author_id(author_data)
            self.summary.loc[i, 'author_id'] = author_id

    def sync_all(self, drop_duplicates = False, drop_empty_rows=False):

        for i in self.all.keys():
            author = self.all[i]
            author.update_id()
            series = author.summary.loc[0]
            all = self.summary.copy(deep=True).astype(str)
            auth_index = all[all['author_id'] == i].index.to_list()[0]
            self.summary.loc[auth_index] = series

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def sync_details(self, drop_duplicates = False, drop_empty_rows=False):

        self.update_author_ids()

        for i in self.summary.index:

            auth_data = self.summary.loc[i]
            auth_id = auth_data['author_id']

            if auth_id != None:
                auth = Author.from_series(auth_data) # type: ignore
                self.all[auth_id] = auth

            else:
                auth_id = generate_author_id(auth_data)
                auth_data['author_id'] = auth_id
                auth = Author.from_series(auth_data) # type: ignore
                self.all[auth_id] = auth
        
        keys = list(self.all.keys())
        for key in keys:
            auth_ids = self.summary['author_id'].to_list()
            if key not in auth_ids:
                del self.all[key]

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def sync(self, drop_duplicates = False, drop_empty_rows=False):
        
        all_len = len(self.summary)
        details_len = len(self.all)

        if all_len > details_len:
            self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
            return
        else:
            if details_len > all_len:
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return
            else:
                self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return


    def drop_empty_rows(self):

        ignore_cols = ['author_id', 'affiliations', 'publications', 'other_links']

        df = self.summary.copy(deep=True)
        df['full_name'] = df['full_name'].replace('no_name_given', None)
        df = df.dropna(axis=0, how='all')
        drop_cols = [c for c in df.columns if c not in ignore_cols]
        df = df.dropna(axis=0, how='all', subset=drop_cols).reset_index().drop('index', axis=1)

        self.summary = df
        self.sync_details()

        return self

    def format_affiliations(self, drop_empty_rows=False):

        if drop_empty_rows == True:
            self.drop_empty_rows()

        affils = self.summary['affiliations'].apply(func=format_affiliations) # type: ignore
        self.summary['affiliations'] = affils
        self.sync_details()

    def update_from_orcid(self, drop_duplicates = False, drop_empty_rows=False):

        self.sync()

        author_ids = self.all.keys()

        for a in author_ids:

            self.all[a].update_from_orcid()
            details = self.all[a].summary.loc[0]
            
            df_index = self.summary[self.summary['author_id'] == a].index.to_list()[0]
            self.summary.loc[df_index] = details

            new_id = details['author_id']
            if new_id != a:
                self.all[new_id] = self.all[a]
                del self.all[a]
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def import_orcid_ids(self, orcid_ids: list, drop_duplicates = False, drop_empty_rows=False):

        for i in orcid_ids:

            auth = Author.from_orcid(i) # type: ignore
            self.add_author(author = auth, data = i)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def from_orcid_ids(orcid_ids: list, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        authors = Authors()
        authors.import_orcid_ids(orcid_ids, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
        
        return authors

    def with_orcid(self):
        return self.summary[~self.summary['orcid'].isna()]

    def import_crossref(self, crossref_result: list, drop_duplicates = False, drop_empty_rows=False):

        for i in crossref_result:

            auth = Author.from_crossref(i) # type: ignore
            self.add_author(author = auth, data = i)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)
    
    def from_crossref(crossref_result: list, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        authors = Authors()
        authors.import_crossref(crossref_result, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)

        return authors
    
    def import_wos(self, wos_result, drop_duplicates = False, drop_empty_rows=False):
        
        authors_data = []

        if type(wos_result) == list:
            authors_data = wos_result
        
        elif type(wos_result) == dict:

            if 'authors' in wos_result.keys():
                authors_data = wos_result['authors']
            else:
                authors_data = []
            
            if 'bookEditors' in wos_result.keys():
                editors_data = wos_result['bookEditors']
            else:
                editors_data = []
        
        authors_df = pd.DataFrame(authors_data)
        authors_df = authors_df.rename(columns={'displayName': 'full_name', 'researcherId': 'wos'}).drop('wosStandard',axis=1)
        
        names_split = authors_df['full_name'].str.split(',').to_list()
        for i in range(0,len(names_split)):
            
            name = names_split[i]

            if len(name) == 0:
                continue

            if len(name) == 1:
                family = name[0]
                full = name[0]
                given = None

            if len(name) > 1:
                given = name[1].strip()
                family = name[0].strip()
                full = given + ' ' + family

            authors_df.loc[i, 'full_name'] = full
            authors_df.loc[i, 'given_name'] = given
            authors_df.loc[i, 'family_name'] = family

        self.summary = pd.concat([self.summary, authors_df])

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.sync_details(drop_duplicates=False, drop_empty_rows=False)

    def from_wos(wos_result, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        authors = Authors()
        authors.import_wos(wos_result=wos_result, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)

        return authors

    def affiliations(self, drop_duplicates = False, drop_empty_rows=False):

        self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
        
        output = {}
        for auth_id in self.all.keys():
            auth = self.all[auth_id]
            affiliation = auth.summary.loc[0, 'affiliations']
            output[auth_id] = affiliation
        
        return output
    
    def search_orcid(self, query: str = 'request_input', add_to_authors: bool = True):

        res = search_orcid(query=query)
        res = res.rename(columns={'credit-name': 'full_name', 'given-names': 'given_name', 'family-name': 'family_name', 'family-names': 'family_name', 'institution-name': 'affiliations', 'orcid-id': 'orcid'}) # type: ignore

        res = res.replace(np.nan, '')

        if add_to_authors == True:

            self.summary = pd.concat([self.summary, res])
            self.update_full_names()
            self.drop_empty_rows()
        
        return res

def format_authors(author_data, drop_duplicates = False, drop_empty_rows=False):
        
        result = Authors()

        if (author_data == None) or (author_data == ''):
            result = Authors()

        if type(author_data) == Authors:
            result = author_data

        if (type(author_data) == list) and (len(author_data) > 0) and (type(author_data[0]) == Author):
            result = Authors()
            result.add_authors_list(author_data)

        if (type(author_data) == list) and (len(author_data) > 0) and (type(author_data[0]) == dict):
            result = Authors.from_crossref(author_data) # type: ignore

        if (type(author_data) == dict) and (len(author_data.values()) > 0) and (type(list(author_data.values())[0]) == str):
            author = Author.from_crossref(author_data) # type: ignore
            result = Authors()
            result.add_author(author)
        
        if (type(author_data) == dict) and ('authors' in author_data.keys()):
            result = Authors.from_wos(wos_result=author_data, drop_duplicates=True, drop_empty_rows=True) # type: ignore
    
        result.format_affiliations()

        if drop_empty_rows == True:
            result.drop_empty_rows()

        if drop_duplicates == True:
            result.remove_duplicates(drop_empty_rows=drop_empty_rows)

        return result







"""Functions for launching internet searches."""

from .webanalysis import open_url, open_url_source

from typing import List, Dict, Tuple
import webbrowser
from urllib.parse import quote
import copy

def search_web(
                query: str = 'request_input',
                search_engine: str = 'Google',
                view_source: bool = False
              ):
    
    """
    Launches web search in the default web browser using user's query.
    
    Parameters
    ----------
    query : str
        query to search. Defaults to requesting from user input.
    search_engine : str
        name of search engine to use. Defaults to 'Google'.
    view_source : bool
        whether to open the page source code of the search result.
    """
    
    # Requesting query from user input if none given
    if query == 'request_input':
        query = input('Search: ')
    
    # Reformatting query
    query = quote(query)
    
    # Creating base for URL depending on search engine specified
    if search_engine == 'Google':
            url_base = 'https://www.google.com/search?hl=en&q='
        
    else:

        if search_engine == 'DuckDuckGo':
            url_base = 'https://duckduckgo.com/?q='

        if search_engine == 'Bing':
            url_base = 'https://www.bing.com/search?q='
    
    # Creating search URL
    url = url_base + query
    
    # If view_source is True, opens the URL's source code
    if view_source == True:
        open_url_source(url = url)
    
    else:
        open_url(url = url)
         
            
def multi_search_web(iteration_terms: list = 'request_input',
                        query: str = 'request_input',
                        search_engine: str = 'Google',
                        view_source: bool = False
                    ):
    
    """
    Launches multiple web searches by iterating on a query through a list of terms.
    
    Parameters
    ----------
    iteration_terms : list, set, or tuple
        iterable of strings to insert into the search query.
    query : str
        query to search. The function replaces '___' with the iteration term. Defaults to requesting from user input.
    search_engine : str
        name of search engine to use. Defaults to 'Google'.
    view_source : bool
        whether to open the page source code of the search result.
    """
    
    # Requesting iteration terms from user input if none given
    if iteration_terms == 'request_input':
        iteration_terms = input('Type search terms to iterate separated by ",": ')
        iteration_terms = iteration_terms.split(',')
        iteration_terms = [i.strip() for i in iteration_terms]
    
    # Requesting query from user input if none given
    if query == 'request_input':
        query = input('Type search query with ___ in place of iteration terms: ')
    
    # Splitting query to input iteration terms
    query_split = query.split('___')
    
    # If only 1 iteration term is provided, runs a normal search
    if len(query_split) == 1:
        return search_web(query = query, 
                   search_engine = search_engine, 
                   view_source = view_source)
    
    print('Search query: ' + query_split[0] + str(iteration_terms) + query_split[1])
    
    
    # Iterating through terms, replacing '___' in query with each term
    for i in iteration_terms:
        
        query_iteration = query_split[0] + str(i) + query_split[1]
        search_web(query = query_iteration, 
                   search_engine = search_engine, 
                   view_source = view_source)

def search_website(query: str = 'request_input', url = 'request_input', view_source = False):
    
    """
    Launches a website-specific Google search for an inputted query and URL.
    
    Parameters
    ----------
    query : str
        query to search. Defaults to requesting from user input.
    url : str
        URL for website to search within.
    view_source : bool
        whether to open the page source code of the search result.
    """
    
    # Requesting query from user input if none given
    if query == 'request_input':
        query = input('Search for: ')
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Formatting query to focus on site
    site_query = 'site:' + url
    query = site_query + ' ' + query
    
    # Opening web search
    search_web(query, search_engine = 'Google', view_source = view_source)
    

def search_social_media(query: str = 'request_input', platform: str = 'Twitter', view_source = False):
    
    """
    Launches a Google search focused on specified social media platform for inputted query.
    
    Parameters
    ----------
    query : str
        query to search. Defaults to requesting from user input.
    platform : str
        name of the social media platform to search. Defaults to 'Twitter'.
    view_source : bool
        whether to open the page source code of the search result.
        
    Notes
    -----
    Options for 'platform':
        * Twitter
        * Facebook
        * Instagram
        * TikTok
        * Threads
        * LinkedIn
        * Reddit
        * 4Chan
    """
    
    # Requesting query from user input if none given
    if query == 'request_input':
        query = input(f'Search {platform} for: ')
    
    # Selecting platform and formatting query
    platform = platform.lower().strip()
    
    if platform == 'twitter':
        site_query = 'site:twitter.com'
    
    if platform == 'facebook':
        site_query = 'site:facebook.com'
    
    if platform == 'instagram':
        site_query = 'site:instagram.com'
    
    if platform == 'tiktok':
        site_query = 'site:tiktok.com'
    
    if platform == 'threads':
        site_query = 'site:threads.net'
        
    if platform == 'linkedin':
        site_query = 'site:linkedin.com'
    
    if platform == 'reddit':
        site_query = 'site:reddit.com'
        
    if platform == '4chan':
        site_query = 'site:4chan.org'
    
    query = site_query + ' ' + query
    
    # Running search
    search_web(query, search_engine = 'Google', view_source = view_source)
        

def search_twitter(query: str = 'request_input'):
    
    """
    Launches a Google search focused Twitter for inputted query.
    """
    
    search_social_media(query = query, platform = 'Twitter')

    
def search_images(query: str = 'request_input',
               search_engine: str = 'Google Images'
              ):
    
    """
    Launches an image search using the default web browser.
    
    Parameters
    ----------
    query : str
        query to search. Defaults to requesting from user input.
    search_engine : str
        name of image search engine to use. Defaults to 'Google Images'.
    """
    
    # Requesting query from user input if none given
    if query == 'request_input':
        query = input(f'Search: ')
    
    # Reformatting query
    query = quote(query)
    
    # Creating base for URL depending on search engine specified
    if search_engine == 'Google Images':
            url_base = 'https://www.google.com/search?tbm=isch&q='
        
    else:

        if search_engine == 'DuckDuckGo':
            url_base = 'https://duckduckgo.com/?q=i!'

        if search_engine == 'Bing':
            url_base = 'https://www.bing.com/images/search?q='
    
    # Creating search URL
    url = url_base + query
    
    # Opening search
    return webbrowser.open(url)
        
        

def reverse_image_search(url = 'request_input',
               search_engine: str = 'Google Images'
              ):
    
    """
    Launches a reverse image search for a URL using the default web browser.
    
    Parameters
    ----------
    url : str
        URL to search. Defaults to requesting from user input.
    search_engine : str
        name of image search engine to use. Defaults to 'Google Images'.
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Creating base for URL depending on search engine specified
    if search_engine == 'Google Images':
        url_base = 'https://lens.google.com/uploadbyurl?url='
        
    else:

        if search_engine == 'Google Lens':
            url_base = 'https://lens.google.com/uploadbyurl?url='
        
        if search_engine == 'Bing':
            url_base = 'https://www.bing.com/images/search?view=detailv2&iss=sbi&form=SBIIRP&sbisrc=UrlPaste&q=imgurl:'
        
        if search_engine == 'TinEye':
            url_base = 'https://tineye.com/search?url='
    
    # Creating URL for search
    url = url_base + url
    
    # Opening search in browser
    return webbrowser.open(url)


from pathlib import Path
from nltk import download # type: ignore
import pandas as pd

# Importing Stopwords corpus as an NLTK text
try:
    from nltk.corpus import stopwords as nltk_stopwords # type: ignore
    nltk_stopwords.words()
except:
    download('stopwords')
    from nltk.corpus import stopwords as nltk_stopwords # type: ignore
    
nltk_stopwords = list(nltk_stopwords.words())

here = str(Path(__file__).parent)

with open(f'{here}/en_stopwords.txt', 'r', encoding='ascii') as file:
    en_stopwords = file.read()
    file.close()
en_stopwords = en_stopwords.replace("'", "").split(', ')

en_stopwords_lower = pd.Series(en_stopwords).str.lower().to_list()
en_stopwords = list(set(en_stopwords_lower + en_stopwords))

with open(f'{here}/html_stopwords.txt', 'r', encoding='utf-8') as file:
    html_stopwords = file.read()
    file.close()
html_stopwords = html_stopwords.replace("'", "").split(', ')

html_stopwords_lower = pd.Series(html_stopwords).str.lower().to_list()
html_stopwords = list(set(html_stopwords_lower + html_stopwords))

combined = list(set(nltk_stopwords + en_stopwords))

stopwords_dict = {
                'nltk': nltk_stopwords,
                'en': en_stopwords,
                'en+nltk': combined,
                'html': html_stopwords
            }

all_stopwords = []
for key in stopwords_dict.keys():
    all_stopwords = all_stopwords + stopwords_dict[key]

stopwords_dict['all'] = all_stopwords


from ..exporters.general_exporters import obj_to_folder, art_class_to_folder

import pandas as pd
import numpy as np

class Entity:

    """
    This is an Entity object. It is designed to store data about an individual entity and their publications.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self):
        
        """
        Initialises entity instance.
        
        Parameters
        ----------
        """


        self.summary = pd.DataFrame(dtype = object)


    
    def __getitem__(self, key):
        
        """
        Retrieves entity attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.summary.columns:
            return self.summary.loc[0, key]

    def get(self, key):
        return self[key]

    def __repr__(self) -> str:
        return str(self.summary.loc[0])

    def search(self, query: str = 'request_input'):

        if query == 'request_input':
            query = input('Search query').strip()
        
        query = query.strip().lower()
        
        self_str = self.summary.copy(deep=True).loc[0].astype(str).str.lower()
        masked = self_str[self_str.str.contains(query)].index
        
        return self.summary.loc[0][masked]
        

    def has_uri(self) -> bool:

        if 'uri' in self.summary.columns:
            uri = self.summary.loc[0, 'uri']

            if (type(uri) == str) and (uri != ''):
                return True
            else:
                return False
        else:
            return False

    def add_dict(self, data: dict):

        if 'name' in data.keys():
            name = data['name']
            self.summary.loc[0, 'name'] = name

        if 'DOI' in data.keys():
            uri = data['DOI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
            self.summary.loc[0, 'uri'] = 'https://doi.org/' + uri
    
    def from_dict(data: dict): # type: ignore

        entity = Entity()
        entity.add_dict(data=data)

        return entity
        
    def add_series(self, series: pd.Series):
        self.summary.loc[0] = series

    def from_series(data: pd.Series): # type: ignore
        entity = Entity()
        entity.add_series(data)

        return entity

    def add_dataframe(self, dataframe: pd.DataFrame):
        series = dataframe.loc[0]
        self.add_series(series) # type: ignore

    def from_dataframe(data: pd.DataFrame): # type: ignore
        entity = Entity()
        entity.add_dataframe(data)

        return entity
    
    def export_folder(self, 
                      folder_name = 'request_input', 
                      folder_address: str = 'request_input', 
                      export_str_as: str = 'txt', 
                      export_dict_as: str = 'json', 
                      export_pandas_as: str = 'csv', 
                      export_network_as: str = 'graphML'
                      ):
        
        art_class_to_folder(obj=self, folder_name=folder_name, folder_address=folder_address, export_str_as=export_str_as, export_dict_as=export_dict_as, export_pandas_as=export_pandas_as, export_network_as=export_network_as)

    
class Entities:

    """
    This is an Entities object. It contains a collection of Entities objects and compiles data about them.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self):
        
        """
        Initialises Entities instance.
        
        Parameters
        ----------
        """

        self.summary = pd.DataFrame(dtype = object)
        

        self.all = dict()

        self.data = []
        

    def __getitem__(self, key):
        
        """
        Retrieves entities attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.all.keys():
            return self.all[key]

        if key in self.summary.columns:
            return self.summary[key]
        
        if key in self.summary.index:
             return self.summary.loc[key]
        
        if (type(key) == int) and (key <= len(self.data)):
            return self.data[key]
    
    def __repr__(self) -> str:

        string = str(self.summary).replace('[','').replace(']','')
        return string
    
    def __len__(self) -> int:
        return len(self.all.keys())

    def export_folder(self, 
                      folder_name = 'request_input', 
                      folder_address: str = 'request_input', 
                      export_str_as: str = 'txt', 
                      export_dict_as: str = 'json', 
                      export_pandas_as: str = 'csv', 
                      export_network_as: str = 'graphML'
                      ):
        
        
        art_class_to_folder(obj=self, folder_name=folder_name, folder_address=folder_address, export_str_as=export_str_as, export_dict_as=export_dict_as, export_pandas_as=export_pandas_as, export_network_as=export_network_as)

    def save_as(self,
                filetype = 'folder',
                file_name = 'request_input', 
                      folder_address: str = 'request_input', 
                      export_str_as: str = 'txt', 
                      export_dict_as: str = 'json', 
                      export_pandas_as: str = 'csv', 
                      export_network_as: str = 'graphML'):
        
        if filetype == 'folder':
            self.export_folder(folder_name=file_name, folder_address=folder_address, export_str_as=export_str_as, export_dict_as=export_dict_as, export_pandas_as=export_pandas_as, export_network_as=export_network_as)


    def drop(self, entity_id):

        if entity_id in self.all.keys():
            del self.all[entity_id]

        id_col = None
        if 'author_id' in self.summary.columns:
            id_col = 'author_id'
        if 'funder_id' in self.summary.columns:
            id_col = 'funder_id'
        if 'affiliation_id' in self.summary.columns:
            id_col = 'affiliation_id'

        i_to_drop = self.summary[self.summary[id_col] == entity_id].index.to_list()

        if len(i_to_drop) > 0:
            self.summary.drop(labels=i_to_drop, axis=0)


    def merge(self, entities):

        left = self.summary.copy(deep=True)
        right = entities.summary.copy(deep=True)
        
        merged = pd.concat([left, right])

        self.summary = merged.drop_duplicates(ignore_index=True)

        for i in entities.all.keys():
            if i not in self.all.keys():
                self.all[i] = entities.all[i]

        left_data = self.data
        right_data = entities.data

        if left_data == None:
                left_data = []
            
        if right_data == None:
                right_data = []

        if (type(left_data) == Entity) or (type(left_data) == str):
                left_data = [left_data]
            
        if (type(right_data) == Entity) or (type(right_data) == str):
                right_data = [right_data]
            
        if type(left_data) == dict:
                left_data = list(left_data.values())
            
        if type(right_data) == dict:
                right_data = list(right_data.values())

        merged_data = left_data + right_data # type: ignore
        merged_data = pd.Series(merged_data).value_counts().index.to_list()

        self.data = merged_data

        return self

    def with_crossref(self):

        if 'crossref_id' in self.summary.columns:
            return self.summary[~self.summary['crossref_id'].isna()]
        else:
            return pd.DataFrame(index=self.summary.columns, dtype=object)
    
    def with_uri(self):

        if 'uri' in self.summary.columns:
            return self.summary[~self.summary['uri'].isna()]
        else:
            return pd.DataFrame(index=self.summary.columns, dtype=object)
    
    def contains(self, query: str = 'request_input', ignore_case: bool = True) -> bool:

        if query == 'request_input':
            query = input('Search query').strip()

        query = query.strip()

        all_str = self.summary.copy(deep=True).astype(str)
        
        if ignore_case == True:
            query = query.lower()
            for c in all_str.columns:
                all_str[c] = all_str[c].str.lower()
            

        cols = self.summary.columns

        for c in cols:

            if '_id' in c:
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if (c == 'name') or (c == 'full_name'):
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
            if c == 'uri':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'orcid':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'google_scholar':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if c == 'scopus':
                res = all_str[all_str[c].str.contains(query)]
                if len(res) > 0:
                    return True
            
            if (c == 'crossref_id') or (c == 'crossref'):
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
            if (c == 'website') or (c == 'link'):
                res = all_str[all_str[c] == query]
                if len(res) > 0:
                    return True
            
        return False
        

    def search_ids(self, query: str = 'request_input'):

        if query == 'request_input':
            query = input('Search query').strip()
        
        query = query.strip().lower()

        cols = [c for c in self.summary.columns if (('_id' in c) or (c == 'uri'))]

        masked_indexes = []
        for col in cols:
                col_str = self.summary[col].copy(deep=True).astype(str).str.lower()
                indexes = col_str[col_str.str.contains(query)].index.to_list()
                masked_indexes = masked_indexes + indexes
        
        masked_indexes = list(set(masked_indexes))
        all_res = self.summary.loc[masked_indexes].copy(deep=True)
        final_indexes = all_res.astype(str).drop_duplicates().index
        result = self.summary.loc[final_indexes]

        return result
        

    def search(self, query: str = 'request_input'):
        
        if query == 'request_input':
            query = input('Search query').strip()
        
        query = query.strip()
        query_list = []

        if 'AND' in query:
            query_list = query.split('AND')
            query_list = [i.strip() for i in query_list]
        else:
            if '&' in query:
                query_list = query.split('&')
                query_list = [i.strip() for i in query_list]

        if len(query_list) > 0:
            
            results_list = []

            for i in query_list:
                res_indexes = set(self.search(query=i).index)
                results_list.append(res_indexes)
            
            intersect = set.intersection(*results_list)
            indexes = list(intersect)

            result = self.summary.loc[indexes]

            return result
            

        
        else:
            query = query.lower()
            masked_indexes = []
            for col in self.summary.columns:
                col_str = self.summary[col].copy(deep=True).astype(str).str.lower()
                indexes = col_str[col_str.str.contains(query)].index.to_list()
                masked_indexes = masked_indexes + indexes
                
            masked_indexes = list(set(masked_indexes))
            all_res = self.summary.loc[masked_indexes].copy(deep=True)

            if 'affiliations' in self.summary.columns:

                affils = self.summary['affiliations']
                
                for i in self.summary.index:
                    
                    a = affils[i]
                    
                    if 'search' in a.__dir__():

                        a_res = a.search(query)
                        val = len(a_res)
                        if val > 0:
                            series = self.summary.loc[i]
                            all_res.loc[i] = series
                
            final_indexes = all_res.astype(str).drop_duplicates().index
            
            result = self.summary.loc[final_indexes]

        return result


                  